#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

// liste chainée pour stocker les voisins d'un sommet
struct adj{
	unsigned int voisin;
	struct adj* suiv;
};
typedef struct adj adj_t;

struct graphe {
	int n;
	adj_t** lv; // listes des voisins		
};
typedef struct graphe graphe_t;

void print_graphe(graphe_t* g){
	printf("Graphe à %d sommets\n", g->n);
	for (int i = 0; i < g->n; i++){
		printf("%d", i);
		if (g->lv[i] != NULL){
			printf(" -> ");
			for (adj_t* v = g->lv[i]; v != NULL; v = v->suiv)
			{
				printf("%d, ", v->voisin);
			}
		}
		printf("\n");
	}
}


void free_graphe(graphe_t* g){
	for (int i = 0; i < g->n; i++){
		adj_t* v = g->lv[i];
		while (v != NULL){
			adj_t* prev = v;
			v = v->suiv;
			free(prev);
		}
	}
	free(g->lv);
	free(g);
}

graphe_t* graphe_vide(int n){
	graphe_t* g = malloc(sizeof(graphe_t));
	g->n = n;
	g->lv = malloc(n*sizeof(adj_t*));
	for (int i = 0; i < n; i++){
		g->lv[i] = NULL;
	}
	return g;
}

void ajouter(graphe_t* g, unsigned int u, unsigned int v, bool oriente){
	// chercher v dans la liste des voisins de u
	adj_t* maillon = g->lv[u];
	while (maillon != NULL && maillon->voisin != v){
		maillon = maillon->suiv;
	}
	// en sortie: soit maillon = NULL, soit maillon->voisin contient v
	// si le maillon est NULL, on rajoute v en tête de la liste
	if (maillon == NULL){
		adj_t* nouv_maillon = malloc(sizeof(adj_t));
		nouv_maillon->voisin = v;
		nouv_maillon->suiv = g->lv[u];
		g->lv[u] = nouv_maillon;
	}

	// si l'arête n'était pas orientée, on rajoute l'arête dans l'autre sens
	if (!oriente){
		ajouter(g, v, u, true);
	}
}


graphe_t* lire_graphe(char* filename){
	int n, m;
	int oriente_lu; // pour pouvoir lire avec %d
	bool oriente;
	graphe_t* g;

	FILE* f = fopen(filename, "r");
	fscanf(f, "%d %d %d", &n, &m, &oriente_lu);
	oriente = oriente_lu;

	g = graphe_vide(n);
	for (int i = 0; i < m; ++i)
	{
		int u, v;
		fscanf(f, "%d %d", &u, &v);
		ajouter(g, u, v, oriente);
	}

	fclose(f);
	return g;
}

/* 
Une remarque: pas besoin du tableau d, on s'en sert pour la preuve 
de correction, mais dans le code pred seul suffit.
*/
unsigned int* bfs(graphe_t* g, unsigned int s){
	unsigned int* pred = malloc(g->n*sizeof(unsigned int));
	for (int i = 0; i < g->n; i++){
		pred[i] = -1; // GCC rale car on stocke -1 dans un tableau d'entiers non signés
	}

	// initialisation de la structure de file
	unsigned int* file = malloc(g->n*sizeof(unsigned int));
	int tete  = 0;
	int queue = 0;

	pred[s] = s;
	file[queue] = s; queue++;
	while (tete != queue){
		int u = file[tete]; tete++;
		for (adj_t* maillon = g->lv[u]; maillon != NULL; maillon = maillon->suiv){
			unsigned int v = maillon->voisin;
			if (pred[v] == -1){
				pred[v] = u;
				file[queue] = v; queue++;
			}
		}

	}
	free(file);
	return pred;
}

/* coquille: int unsigned -> unsigned int */
unsigned int* plus_court_chemin(graphe_t* g, unsigned int s, unsigned t){
	unsigned int* pred = bfs(g, s);
	if (pred[t] == -1){
		printf("Pas de chemin\n");
		return NULL;
	}

	// reconstruire le résultat à l'envers
	unsigned int* res = malloc(g->n*sizeof(unsigned int));
	res[0] = t;
	unsigned int u = t;
	int k = 1; // indice d'écriture dans res
	while (u != s){
		u = pred[u];
		res[k] = u;
		k++;
	}

	// renverser puis renvoyer:
	//      échanger les cases 0 et k-1
	// puis échanger les cases 1 et k-2
	// ...
	// puis échanger les cases partie entière (k/2) et partie supérieure(k/2)
	for (int i = 0; i<k/2;i++){
		int tmp = res[i];
		res[i] = res[k-1-i];
		res[k-1-i] = tmp;
	}

	free(pred);
	return res;
}


int main(int argc, char const *argv[])
{	
	graphe_t* g = malloc(sizeof(graphe_t));
	g->n = 4;
	g->lv = malloc(4*sizeof(adj_t*));
	
	// successeurs de 0: [1]
	g->lv[0] = malloc(sizeof(adj_t));
	g->lv[0]->voisin = 1;
	g->lv[0]->suiv = NULL;

	// successeurs de 1: [3]
	g->lv[1] = malloc(sizeof(adj_t));
	g->lv[1]->voisin = 3;
	g->lv[1]->suiv = NULL;

	//successeurs de 2: [0, 1, 3]
	g->lv[2] = malloc(sizeof(adj_t));
	g->lv[2]->voisin = 0;
	g->lv[2]->suiv = malloc(sizeof(adj_t));
	g->lv[2]->suiv->voisin = 1;
	g->lv[2]->suiv->suiv = malloc(sizeof(adj_t));
	g->lv[2]->suiv->suiv->voisin = 3;
	g->lv[2]->suiv->suiv->suiv = NULL;

	// successeurs de 3: [0]
	g->lv[3] = malloc(sizeof(adj_t));
	g->lv[3]->voisin = 0;
	g->lv[3]->suiv = NULL;


	print_graphe(g);
	free_graphe(g);

	g = lire_graphe("graphe_test.txt");

	print_graphe(g);

	// chercher et afficher un pcc de 0 à 8 dans le graphe en exemple
	unsigned int* chemin = plus_court_chemin(g, 0, 8);
	for (int i = 0; chemin[i] != 8; i++){
		printf("%d, ", chemin[i]);
	}
	printf("%d\n", 8);
	

	free(chemin);
	free_graphe(g);
	return 0;
}